@@ -15,8 +15,8 @@ our @EXPORT_OK = qw(wrap_sub);
 
 our $Log_Wrapper_Code = $ENV{LOG_PERINCI_WRAPPER_CODE} // 0;
 
-our $VERSION = '0.62'; # VERSION
-our $DATE = '2014-09-06'; # DATE
+our $VERSION = '0.63'; # VERSION
+our $DATE = '2014-10-15'; # DATE
 
 our %SPEC;
 
@@ -526,25 +526,27 @@ sub _plc {
     state $plc = $self->_sah->get_compiler("perl");
 }
 
-sub handlemeta_args { {v=>2, prio=>10} }
-sub handle_args {
+sub _handle_args {
     my ($self, %args) = @_;
 
-    my $v = $self->{_meta}{args};
+    my $v = $args{v} // $self->{_meta}{args};
     return unless $v;
 
     my $opt_sin = $self->{_args}{_schema_is_normalized};
     my $opt_va  = $self->{_args}{validate_args};
 
+    my $prefix = $args{prefix} // '';
+    my $argsterm = $args{argsterm} // '%args';
+
     if ($opt_va) {
         $self->_add_module("experimental 'smartmatch'");
         $self->select_section('before_call_arg_validation');
-        $self->push_lines('', '# check args');
-        $self->push_lines('for (sort keys %args) {');
+        $self->push_lines('', '# check args') if $prefix eq '';
+        $self->push_lines("for (sort keys $argsterm) {");
         $self->indent;
-        $self->_errif(400, q["Invalid argument name '$_'"],
+        $self->_errif(400, q["Invalid argument name '].$prefix.q[$_'"],
                       '!/\A(-?)\w+(\.\w+)*\z/o');
-        $self->_errif(400, q["Unknown argument '$_'"],
+        $self->_errif(400, q["Unknown argument '].$prefix.q[$_'"],
                       '!($1 || $_ ~~ '.__squote([sort keys %$v]).')');
         $self->unindent;
         $self->push_lines('}');
@@ -553,7 +555,14 @@ sub handle_args {
     for my $argname (sort keys %$v) {
         my $argspec = $v->{$argname};
 
-        my $argterm = "\$args{'$argname'}";
+        my $argterm = $argsterm;
+        if ($argterm =~ /^%\{\s*(.+)\s*\}/) {
+            $argterm = $1 . "->{'$argname'}";
+        } elsif ($argterm =~ s/^%/\$/) {
+            $argterm .= "{'$argname'}";
+        } else {
+            $argterm .= "->{'$argname'}";
+        }
 
         my $sch = $argspec->{schema};
         if ($sch) {
@@ -566,7 +575,7 @@ sub handle_args {
                     data_name            => $dn,
                     data_term            => $argterm,
                     schema               => $sch,
-                    schema_is_normalized => !$opt_sin,
+                    schema_is_normalized => $opt_sin,
                     return_type          => 'str',
                     indent_level         => $self->get_indent_level + 4,
                     %{ $self->{_args}{_extra_sah_compiler_args} // {}},
@@ -578,8 +587,36 @@ sub handle_args {
                 $self->indent;
                 $self->push_lines("my \$err_$dn;\n$cd->{result};");
                 $self->_errif(
-                    400, qq["Invalid value for argument '$argname': \$err_$dn"],
+                    400, qq["Invalid value for argument '$prefix$argname': \$err_$dn"],
                     "\$err_$dn");
+                if ($argspec->{meta}) {
+                    $self->push_lines("# check subargs of $prefix$argname");
+                    $self->_handle_args(
+                        %args,
+                        v => $argspec->{meta}{args},
+                        prefix => ($prefix ? "$prefix/" : "") . "$argname/",
+                        argsterm => '%{'.$argterm.'}',
+                    );
+                }
+                if ($argspec->{element_meta}) {
+                    $self->push_lines("# check element subargs of $prefix$argname");
+                    my $indexterm = "$prefix$argname";
+                    $indexterm =~ s/\W+/_/g;
+                    $indexterm = '$i_' . $indexterm;
+                    $self->push_lines('for my '.$indexterm.' (0..$#{ '.$argterm.' }) {');
+                    $self->indent;
+                    $self->_errif(
+                        400, qq("Invalid value for argument '$prefix$argname\[).qq($indexterm]': must be hash"),
+                        "ref($argterm\->[$indexterm]) ne 'HASH'");
+                    $self->_handle_args(
+                        %args,
+                        v => $argspec->{element_meta}{args},
+                        prefix => ($prefix ? "$prefix/" : "") . "$argname\[$indexterm]/",
+                        argsterm => '%{'.$argterm.'->['.$indexterm.']}',
+                    );
+                    $self->unindent;
+                    $self->push_lines('}');
+                }
                 $self->unindent;
                 if ($has_default_prop) {
                     $self->push_lines(
@@ -601,6 +638,12 @@ sub handle_args {
     }
 }
 
+sub handlemeta_args { {v=>2, prio=>10} }
+sub handle_args {
+    my ($self, %args) = @_;
+    $self->_handle_args(%args);
+}
+
 sub handlemeta_result { {v=>2, prio=>50} }
 sub handle_result {
     require Data::Sah;
@@ -1197,7 +1240,7 @@ Perinci::Sub::Wrapper - A multi-purpose subroutine wrapping framework
 
 =head1 VERSION
 
-This document describes version 0.62 of Perinci::Sub::Wrapper (from Perl distribution Perinci-Sub-Wrapper), released on 2014-09-06.
+This document describes version 0.63 of Perinci::Sub::Wrapper (from Perl distribution Perinci-Sub-Wrapper), released on 2014-10-15.
 
 =head1 SYNOPSIS
 
@@ -1375,11 +1418,11 @@ The wrapped subroutine along with its new metadata (hash)
 
 Aside from wrapping the subroutine, the wrapper will also create a new metadata
 for the subroutine. The new metadata is a clone of the original, with some
-properties changed, e.g. schema in C<args> and C<result> normalized, some values
-changed according to the C<convert> argument, some defaults set, etc.
+properties changed, e.g. schema in `args` and `result` normalized, some values
+changed according to the `convert` argument, some defaults set, etc.
 
 The new metadata will also contain (or append) the wrapping log located in the
-C<x.perinci.sub.wrapper.logs> attribute. The wrapping log marks that the wrapper
+`x.perinci.sub.wrapper.logs` attribute. The wrapping log marks that the wrapper
 has added some functionality (like validating arguments or result) so that
 future nested wrapper can choose to avoid duplicating the same functionality.
 
@@ -1554,7 +1597,7 @@ Please visit the project's homepage at L<https://metacpan.org/release/Perinci-Su
 
 =head1 SOURCE
 
-Source repository is at L<https://github.com/sharyanto/perl-Perinci-Sub-Wrapper>.
+Source repository is at L<https://github.com/perlancar/perl-Perinci-Sub-Wrapper>.
 
 =head1 BUGS
 
